/************
*
*   This file is part of a tool for producing 3D content in the PRC format.
*   Copyright (C) 2008  Orest Shardt <shardtor (at) gmail dot com>
*
*   This program is free software: you can redistribute it and/or modify
*   it under the terms of the GNU Lesser General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU Lesser General Public License for more details.
*
*   You should have received a copy of the GNU Lesser General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.

*
*************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <string>
#include <vector>
#include <map>
#include "oPRCFile.h"
#include <ticpp.h>
#include <stdlib.h>
#include <setjmp.h>
#include <math.h>
#include <hpdf.h>
#include <hpdf_conf.h>
#include <hpdf_u3d.h>
#include <hpdf_annotation.h>

HPDF_Dict HPDF_LoadJSFromFile  (HPDF_Doc     pdf, const char  *filename)
{
	HPDF_Stream js_data;
	HPDF_Dict js = NULL;

//	HPDF_PTRACE ((" HPDF_LoadJSFromFile\n"));

	if (!HPDF_HasDoc (pdf)) {
		return NULL;
	}

	/* create file stream */
	js_data = HPDF_FileReader_New (pdf->mmgr, filename);

	if (!HPDF_Stream_Validate (js_data)) {
		HPDF_Stream_Free (js_data);
		return NULL;
	}

	js = HPDF_DictStream_New (pdf->mmgr, pdf->xref);
	if (!js) {
		HPDF_Stream_Free (js_data);
		return NULL;
	}

	js->header.obj_class |= HPDF_OSUBCLASS_XOBJECT;

	/* add required elements */
	js->filter = HPDF_STREAM_FILTER_NONE;

	for (;;) {
		HPDF_BYTE buf[HPDF_STREAM_BUF_SIZ];
		HPDF_UINT len = HPDF_STREAM_BUF_SIZ;
		HPDF_STATUS ret = HPDF_Stream_Read (js_data, buf, &len);

		if (ret != HPDF_OK) {
			if (ret == HPDF_STREAM_EOF) {
				if (len > 0) {
					ret = HPDF_Stream_Write (js->stream, buf, len);
					if (ret != HPDF_OK) {
						HPDF_Stream_Free (js_data);
						HPDF_Dict_Free(js);
						return NULL;
					}
				}
				break;
			} else {
				HPDF_Stream_Free (js_data);
				HPDF_Dict_Free(js);
				return NULL;
			}
		}

		if (HPDF_Stream_Write (js->stream, buf, len) != HPDF_OK) {
			HPDF_Stream_Free (js_data);
			HPDF_Dict_Free(js);
			return NULL;
		}
	}

	/* destroy file stream */
	HPDF_Stream_Free (js_data);

	if (!js) {
		HPDF_CheckError (&pdf->error);
	}
	return js;
}


PRCRepresentationItemList transparentMeshes; // there is some bug - transparency for vertex colored meshes can be set via JS or PDF dictionary, but not inside model file (in OpenGL rendering mode)
std::vector<double> transparencies;
double mat[3][3]; // viewpoint transform
uint32_t black_material_generic_index = m1; // base material for textures

using namespace std;
using namespace ticpp;

struct proto
{
   uint32_t tess_index;
   string type;
};

map<string, proto> protos;

struct X3DColor
{
  X3DColor() : r(0),g(0),b(0) {}
  X3DColor(double R, double G, double B) : r(R),g(G),b(B) {}
  friend ostream& operator<<(ostream& output, const X3DColor& col);
  friend istream& operator>>(istream& input, X3DColor& col);
  double r;
  double g;
  double b;
  bool operator==(const X3DColor &c) const
  {
   return r==c.r && g==c.g && b==c.b;
  }
};

ostream& operator<<(ostream& output, const X3DColor& col)
{
    output << col.r << ',' << col.g << ',' << col.b;
    return output;
}

istream& operator>>(istream& input, X3DColor& col)
{
  input >> col.r >> col.g >> col.b;
  return input;
}

// disabled since Adobe Acrobat/Reader do not support linetype and Maple does not output linewidth
// struct X3DLineProperties
// {
//   X3DLineProperties() : linetype(1),linewidthScaleFactor(0) {}
//   friend ostream& operator<<(ostream& output, const X3DLineProperties& line);
//   int32_t linetype;
//   double linewidthScaleFactor;
//   bool operator==(const X3DLineProperties &line) const
//   {
//    return linetype==line.linetype && linewidthScaleFactor==line.linewidthScaleFactor;
//   }
// };
// 
// ostream& operator<<(ostream& output, const X3DLineProperties& line)
// {
//     output << "linetype: " << line.linetype << " linewidthScaleFactor: " << line.linewidthScaleFactor;
//     return output;
// }

struct X3DMaterial
{
  X3DMaterial() : ambientIntensity(0.2), diffuseColor(0.8,0.8,0.8), emissiveColor(0,0,0), shininess(0.2), specularColor(0,0,0), transparency(0) {}
  double ambientIntensity;
  X3DColor diffuseColor;
  X3DColor emissiveColor;
  double shininess;
  X3DColor specularColor;
  double transparency;
  bool operator==(const X3DMaterial &m) const
  {
   return ambientIntensity==m.ambientIntensity && diffuseColor==m.diffuseColor && emissiveColor==m.emissiveColor && shininess==m.shininess && specularColor==m.specularColor && transparency==m.transparency;
  }
  friend ostream& operator<<(ostream& output, const X3DMaterial& col);
};

ostream& operator<<(ostream& output, const X3DMaterial& mat)
{
    output << "diffuseColor: " << mat.diffuseColor << endl;
    output << "emissiveColor: " << mat.emissiveColor << endl;
    output << "specularColor: " << mat.specularColor << endl;
    output << "ambientIntensity: " << mat.ambientIntensity << endl;
    output << "shininess: " << mat.shininess << endl;
    output << "transparency: " << mat.transparency << endl;
    return output;
}

struct X3DIndexedTriangleSet
{
  X3DIndexedTriangleSet() : solid(true) {}
  bool solid;
  vector<int32_t> index;
  vector<X3DColor> color;
  vector<PRCVector3d> point;
};

struct X3DTriangleFanSet
{
  X3DTriangleFanSet() : solid(true) {}
  bool solid;
  vector<int32_t> fanCount;
  vector<X3DColor> color;
  vector<PRCVector3d> point;
};

struct X3DIndexedLineSet
{
  vector<int32_t> index;
  vector<X3DColor> color;
  vector<PRCVector3d> point;
};

struct X3DRotation
{
  X3DRotation() : axis(0,0,1), angle(0) {}
  PRCVector3d axis;
  double angle;
 bool operator==(const X3DRotation &r) const
 {
  return axis==r.axis && angle==r.angle;
 }

};

struct X3DTransform
{
  X3DTransform() : center(0,0,0), scale(1,1,1), translation(0,0,0) {}
  friend ostream& operator<<(ostream& output, const X3DTransform& trans);
  PRCVector3d center;
  X3DRotation rotation;
  PRCVector3d scale;
  X3DRotation scaleOrientation;
  PRCVector3d translation;
  bool operator==(const X3DTransform &t) const
  {
   return center==t.center && rotation==t.rotation && scale==t.scale && scaleOrientation==t.scaleOrientation && translation==t.translation;
  }
};
typedef vector<X3DTransform> X3DTransformList;

ostream& operator<<(ostream& output, const X3DTransform& trans)
{
  if(!(trans.center==PRCVector3d(0,0,0)))
    output << "center: " << trans.center.x << ',' << trans.center.y << ',' << trans.center.z << ' ';
  if(!(trans.rotation==X3DRotation()))
    output << "rotation: " << trans.rotation.axis.x << ',' << trans.rotation.axis.y << ',' << trans.rotation.axis.z << ' ' << trans.rotation.angle;
  if(!(trans.scale==PRCVector3d(1,1,1)))
    output << "scale: " << trans.scale.x << ',' << trans.scale.y << ',' << trans.scale.z << ' ';
  if(!(trans.scaleOrientation==X3DRotation()))
    output << "scaleOrientation: " << trans.scaleOrientation.axis.x << ',' << trans.scaleOrientation.axis.y << ',' << trans.scaleOrientation.axis.z << ' ' << trans.scaleOrientation.angle << ' ';
  if(!(trans.translation==PRCVector3d(0,0,0)))
    output << "translation: " << trans.translation.x << ',' << trans.translation.y << ',' << trans.translation.z << ' ';
    output << endl;
  return output;
}


void processElement(ticpp::Element* pElem, X3DTransformList transforms, string DEF, oPRCFile &file, PRCPartDefinition *part_definition);
void readIndexedTriangleSet(X3DIndexedTriangleSet &set, ticpp::Element *pSet);
void readTriangleFanSet(X3DTriangleFanSet &set, ticpp::Element *pSet);
void readIndexedLineSet(X3DIndexedLineSet &set, ticpp::Element *pSet);
uint32_t maketessIndexedTriangleStripSet(const X3DIndexedTriangleSet &set, oPRCFile &file);
uint32_t maketessIndexedTriangleFanSet(const X3DIndexedTriangleSet &set, oPRCFile &file);
uint32_t maketessTriangleFanSet(const X3DTriangleFanSet &set, oPRCFile &file);
uint32_t maketessIndexedLineSet(const X3DIndexedLineSet &set, oPRCFile &file);

int main(int argc, char *argv[])
{
  const char *egog = "the only argument to the program must be an X3D file exported from Maple as Extensible 3D";
  if(argc != 2)
  {
    fputs(egog,stderr);
    exit(1);
  }
  const char *x3dfile = argv[1];
#ifndef PATH_MAX
#define PATH_MAX 256
#endif
  char fname[PATH_MAX]; // file name without extension
  memset( fname, 0,  sizeof(fname) );
  strncpy(fname, x3dfile, sizeof(fname)-1);
  const size_t fnlen = strlen(fname);
#ifdef _MSC_VER
#define strncasecmp _strnicmp
#endif
  if ( fnlen < 5 || strncasecmp(fname+fnlen-4, ".x3d",4) != 0 )
  {
    fputs(egog,stderr);
    exit(1);
  }

  fname[fnlen-4]='\0';
  strcat(fname,".prc");
  oPRCFile file(fname);

  PRCPartDefinition *part_definition = new PRCPartDefinition;

  // first load the xml file
  ticpp::Document doc(argv[1]);
  doc.LoadFile();
/*
  // parse through all elements
  ticpp::Iterator<ticpp::Element> child;
  for(child = child.begin(doc.FirstChildElement()); child != child.end(); child++)
  {
    // The value of this child identifies the name of this element
    string strName;
    string strValue;
    child->GetValue(&strName);
    cout << "child: " << strName << endl;
    cout << "-------------" << endl;
  
    // now parse through all the attributes of this fruit
    Iterator< ticpp::Attribute > attribute;
    for(attribute = attribute.begin(child.Get()); attribute != attribute.end(); attribute++)
      {
              attribute->GetName(&strName);
              attribute->GetValue(&strValue);
              cout << strName << ": " << strValue << endl;
      }
      cout << endl;
  } */

  ticpp::Element* pRoot;
  pRoot=doc.FirstChildElement();
  ticpp::Element* pScene;
  pScene=doc.FirstChildElement()->FirstChildElement( "Scene" );

  X3DColor background;
  X3DTransformList transforms;

  // parse through all elements
  ticpp::Iterator<ticpp::Element> child;
  for(child = child.begin(pScene); child != child.end(); child++)
  {
    string strName;
    string strValue;

    child->GetValue(&strName);
    if(strName=="Background")
    {
      child->GetAttribute( "skyColor", &strValue);
      std::istringstream stringstream(strValue);
      stringstream >> background.r >> background.g >> background.b;
//    cout << "background" << ": " << background << endl;
    }

    if(strName=="ProtoDeclare")
    {
      child->GetAttribute( "name", &strName);
      ticpp::Element* pShape;
      pShape = child->FirstChildElement( "ProtoBody" )->FirstChildElement( "Shape" );

      uint32_t tess_index = m1;
      if(pShape->FirstChildElement( "IndexedTriangleStripSet", false ))
      {
        X3DIndexedTriangleSet set;
        readIndexedTriangleSet(set, pShape->FirstChildElement( "IndexedTriangleStripSet" ));
        tess_index = maketessIndexedTriangleStripSet(set, file);
        protos[strName].type = "IndexedTriangleSet";
      }
      if(pShape->FirstChildElement( "IndexedTriangleFanSet", false ))
      {
        X3DIndexedTriangleSet set;
        readIndexedTriangleSet(set, pShape->FirstChildElement( "IndexedTriangleFanSet" ));
        tess_index = maketessIndexedTriangleFanSet(set, file);
        protos[strName].type = "IndexedTriangleSet";
      }
      if(pShape->FirstChildElement( "TriangleFanSet", false ))
      {
        X3DTriangleFanSet set;
        readTriangleFanSet(set, pShape->FirstChildElement( "TriangleFanSet" ));
        tess_index = maketessTriangleFanSet(set, file);
        protos[strName].type = "TriangleFanSet";
      }
      if(pShape->FirstChildElement( "IndexedLineSet", false ))
      {
        X3DIndexedLineSet set;
        readIndexedLineSet(set, pShape->FirstChildElement( "IndexedLineSet" ));
        tess_index = maketessIndexedLineSet(set, file);
        protos[strName].type = "IndexedLineSet";
      }
      protos[strName].tess_index = tess_index;
    }

    if(strName=="Transform"||strName=="Group")
      processElement(child.Get(), transforms, "", file, part_definition);

  }

  PRCProductOccurrence *product_occurrence = new PRCProductOccurrence("root");
  product_occurrence->index_part = file.addPartDefinition(part_definition);


    HPDF_Rect rect = {0, 0, 600, 600};

    HPDF_Doc  pdf;
    HPDF_Page page;
    HPDF_Annotation annot;
    HPDF_U3D u3d;

    pdf = HPDF_New (NULL, NULL);

    pdf->pdf_version = HPDF_VER_17;

    page = HPDF_AddPage (pdf);

    HPDF_Page_SetWidth (page, 600);
    HPDF_Page_SetHeight (page, 600);

    HPDF_Dict js = NULL;

    js = HPDF_DictStream_New (pdf->mmgr, pdf->xref);

    js->header.obj_class |= HPDF_OSUBCLASS_XOBJECT;

    /* add required elements */
    js->filter = HPDF_STREAM_FILTER_NONE;

    const unsigned char script[] = 
"\
// arrays of specially-named sections of the model tree \n\
var BBOARDobjsArray  = new Array(); // stores the billboard meshes \n\
var BBOARDposArray   = new Array(); // stores the billboard positions \n\
var BBOARDscaleArray = new Array(); // stores the billboard scales \n\
var BBOARDCount = 0;           // how many billboard meshes are there? \n\
var PrettyDrawingObjsArray      = new Array(); \n\
var PrettyDrawingPosArray       = new Array(); \n\
var PrettyDrawingTransformArray = new Array(); \n\
var PrettyDrawingCount = 0; \n\
 \n\
// temporary variables \n\
var currMesh = null; \n\
var currname = \"\"; \n\
 \n\
// compile lists of BBOARD*, VRSET* and FRAME* meshes \n\
for (i=0; i<scene.meshes.count; i++) {  \n\
  currMesh = null; \n\
  currname = \"\"; \n\
  currMesh = scene.meshes.getByIndex(i);  \n\
  currname = currMesh.name; \n\
 \n\
  if ( currname.indexOf(\"Billboard\") > -1 ) { \n\
    BBOARDobjsArray[BBOARDCount] = currMesh; \n\
    var bb_scale = Math.pow(Math.abs(currMesh.transform.determinant),1.0/3.0);	// scale. \n\
    var bb_center = currMesh.computeBoundingBox().center; 	// center (vector3) \n\
    BBOARDposArray[BBOARDCount] = currMesh.transform.translation.addScaled(bb_center, bb_scale); 	// center position (vector3) \n\
    BBOARDscaleArray[BBOARDCount] = Matrix4x4().translate(bb_center.scale(-1)).scale(bb_scale, bb_scale, bb_scale); \n\
    BBOARDCount++; \n\
  } \n\
  if ( currname.indexOf(\"LinePrettyDrawing\") > -1 ) { \n\
    PrettyDrawingObjsArray[PrettyDrawingCount] = currMesh; \n\
    PrettyDrawingPosArray[PrettyDrawingCount] = currMesh.transform.translation.add(currMesh.computeBoundingBox().center); 	// center position (vector3) \n\
    PrettyDrawingTransformArray[PrettyDrawingCount] = Matrix4x4(currMesh.transform); \n\
    PrettyDrawingCount++; \n\
  } \n\
 \n\
 } \n\
 \n\
// Get Camera \n\
var camera = scene.cameras.getByIndex(0); \n\
 \n\
function doBillboard() { \n\
	 \n\
	// Loop through all billboards, orientating their view direction with the camera position. \n\
	var cam_posn = camera.position.subtract(camera.targetPosition);	// Camera's position vector (vector3) \n\
	var cam_up = camera.up.subtract(camera.position);		// Camera's up vector (vector3) \n\
	for (j = 0; j < BBOARDCount; j++) { \n\
		var bb_posn = BBOARDposArray[j]; \n\
		var bb_trans = Matrix4x4(); \n\
		bb_trans.setView(bb_posn, cam_posn.add(bb_posn), cam_up.add(bb_posn)); \n\
		BBOARDobjsArray[j].transform.set(BBOARDscaleArray[j].multiply(bb_trans));	// Set billboard view. \n\
	} \n\
 \n\
} \n\
 \n\
function doPrettyDrawing() { \n\
	 \n\
	// Loop through all PrettyDrawings shifting them slighly towards the camera position. \n\
	var cam_posn = camera.position;	// Camera's position vector (vector3) \n\
	for (j = 0; j < PrettyDrawingCount; j++) { \n\
		var bb_posn = PrettyDrawingPosArray[j]; \n\
		PrettyDrawingObjsArray[j].transform.set(PrettyDrawingTransformArray[j].translate(camera.position.subtract(PrettyDrawingPosArray[j]).scale(0.0001))); \n\
	} \n\
 \n\
} \n\
 \n\
mreh = new MouseEventHandler(); \n\
mreh.onMouseDown = false; \n\
mreh.onMouseMove = true; \n\
mreh.reportAllTargets = false; \n\
mreh.onEvent = function(event) { \n\
  if (event.leftButtonDown) { \n\
    doBillboard(); \n\
    doPrettyDrawing(); \n\
  } \n\
} \n\
runtime.addEventHandler(mreh); \n\
 \n\
ceh = new RenderEventHandler(); \n\
ceh.onEvent = function(event) { \n\
  doBillboard(); \n\
  doPrettyDrawing(); \n\
} \n\
runtime.addEventHandler(ceh); \n\
 \n\
doBillboard(); \n\
doPrettyDrawing(); \n\
 \n\
// variable part \n\
";

  ofstream jsfile;
  fname[fnlen-4]='\0';
  strcat(fname,".js");
  jsfile.open (fname);
  jsfile << script;
  // set camera
  jsfile << "camera.projectionType = \"perspective\";\n";
  jsfile << "camera.fov = 0.034987357;\n";
  jsfile << "camera.position.set(" << 1000*mat[0][2] << ',' << 1000*mat[1][2] << ',' << 1000*mat[2][2] << ");\n";
  jsfile << "camera.up.set("  << 1000*mat[0][1] << ',' << 1000*mat[1][1] << ',' << 1000*mat[2][1] << ");\n";
  jsfile << "scene.background.setColor(new Color(" << background << "));\n";

  // workaround for vertex color disabling transparency in OpenGL
  for(size_t i=0; i<transparentMeshes.size(); i++)
  {
    uint8_t *serialization_buffer = NULL;
    PRCbitStream serialization(serialization_buffer,0u);
    const PRCFileStructure *pfile_structure = file.fileStructures[0];
    const PRCRepresentationItem *pItem = transparentMeshes[i];
    const uint32_t *unique_id = pfile_structure->header.fileStructureUUID;
 // ConvertUniqueIdentifierToString (prc_entity_unique_id)
 // SerializeCompressedUniqueId (file_structure)
    serialization << unique_id[0] << unique_id[1] << unique_id[2] << unique_id[3];
 // WriteUnsignedInteger (type)
    serialization << pItem->getType();
 // WriteUnsignedInteger (unique_identifier)
    serialization << pItem->getPRCID();
 // serialization_buffer = Flush serialization (serialization) 
    {
       const uint32_t size_serialization = serialization.getSize();
       while(size_serialization == serialization.getSize())
         serialization << false;
    }
  // ConvertUniqueIdentifierToString (prc_occurrence_unique_id)
 // SerializeCompressedUniqueId (file_structure)
    serialization << unique_id[0] << unique_id[1] << unique_id[2] << unique_id[3];
 // WriteUnsignedInteger (type)
    serialization << (uint32_t)PRC_TYPE_ASM_ProductOccurence;
 // WriteUnsignedInteger (unique_identifier)
    serialization << product_occurrence->getPRCID();
 
    jsfile << "scene.meshes.getByName(\"" << pItem->name << '.';
    const uint32_t size_serialization = serialization.getSize();
    for(size_t j=0; j<size_serialization; j++)
      jsfile << hex << setfill('0') << setw(2) << (uint32_t)(serialization_buffer[j]);
    jsfile << "\").renderMode = \"transparent\";" << "\n";
    jsfile << "scene.meshes.getByName(\"" << pItem->name << '.';
    for(size_t j=0; j<size_serialization; j++)
      jsfile << hex << setfill('0') << setw(2) << (uint32_t)(serialization_buffer[j]);
    jsfile << "\").material.opacity = " << 1-transparencies[i] << "; // does not work\n" << "\n";
  }

    jsfile.close();
    js = HPDF_LoadJSFromFile  (pdf, fname);

  file.addProductOccurrence(product_occurrence);
  file.finish();

    fname[fnlen-4]='\0';
    strcat(fname,".prc");
    u3d = HPDF_LoadU3DFromFile (pdf, fname);

    HPDF_Dict_Add (u3d, "OnInstantiate", js);

    //  Create annotation
    annot = HPDF_Page_Create3DAnnot (page, rect, u3d );

    /* save the document to a file */
    fname[fnlen-4]='\0';
    strcat(fname,".pdf");
    HPDF_SaveToFile (pdf, fname);

    /* clean up */
    HPDF_Free (pdf);


  return 0;
}

void readIndexedTriangleSet(X3DIndexedTriangleSet &set, ticpp::Element *pSet)
{
  string strValue;
  pSet->GetAttribute( "solid", &strValue, false);
  if(strValue=="false")
    set.solid = false;
  pSet->GetAttribute( "index", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      int32_t index;
      stringstream >> index;
      stringstream >> ws;
      set.index.push_back(index);
    }
  }
  if(pSet->FirstChildElement( "Color", false ))
  {
    pSet->FirstChildElement( "Color" )->GetAttribute( "color", &strValue);
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      X3DColor color;
      stringstream >> color.r >> color.g >> color.b;
      stringstream >> c >> ws;
      set.color.push_back(color);
    }
  }
  pSet->FirstChildElement( "Coordinate" )->GetAttribute( "point", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      PRCVector3d point;
      stringstream >> point.x >> point.y >> point.z;
      stringstream >> c >> ws;
      set.point.push_back(point);
    }
  }
}

uint32_t maketessIndexedTriangleStripSet(const X3DIndexedTriangleSet &set, oPRCFile &file)
{
  PRC3DTess *tess = new PRC3DTess;
  for(size_t i = 0; i < set.point.size(); ++i)
  {
    tess->coordinates.push_back(set.point[i].x);
    tess->coordinates.push_back(set.point[i].y);
    tess->coordinates.push_back(set.point[i].z);
  }
  PRCTessFace *tessFace = new PRCTessFace;
  tessFace->used_entities_flag=PRC_FACETESSDATA_TriangleStripe;
  uint32_t numberOfTriangleStripes = 0;
  uint32_t numberPoints = 0;
  tessFace->sizes_triangulated.push_back(numberOfTriangleStripes); // to be filled later
  for(size_t i = 0; i < set.index.size(); ++i)
  {
    if(set.index[i]==-1)
    {
      tessFace->sizes_triangulated.push_back(numberPoints);
      numberPoints = 0;
      numberOfTriangleStripes++;
    }
    else
    {
      tess->triangulated_index.push_back(set.index[i]*3);
      if(!set.color.empty())
      {
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].r*255));
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].g*255));
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].b*255));
      }
      numberPoints++;
    }
  }
  if(!set.color.empty())
    tessFace->is_rgba=false;

  tessFace->sizes_triangulated[0]=numberOfTriangleStripes;

  tess->face_tessellation.push_back(tessFace);
  return file.add3DTess(tess);
}

uint32_t maketessIndexedTriangleFanSet(const X3DIndexedTriangleSet &set, oPRCFile &file)
{
  PRC3DTess *tess = new PRC3DTess;
  for(size_t i = 0; i < set.point.size(); ++i)
  {
    tess->coordinates.push_back(set.point[i].x);
    tess->coordinates.push_back(set.point[i].y);
    tess->coordinates.push_back(set.point[i].z);
  }
  PRCTessFace *tessFace = new PRCTessFace;
  tessFace->used_entities_flag=PRC_FACETESSDATA_TriangleFan;
  uint32_t numberOfTriangleFans = 0;
  uint32_t numberPoints = 0;
  tessFace->sizes_triangulated.push_back(numberOfTriangleFans); // to be filled later
  for(size_t i = 0; i < set.index.size(); ++i)
  {
    if(set.index[i]==-1)
    {
      tessFace->sizes_triangulated.push_back(numberPoints);
      numberPoints = 0;
      numberOfTriangleFans++;
    }
    else
    {
      tess->triangulated_index.push_back(set.index[i]*3);
      if(!set.color.empty())
      {
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].r*255));
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].g*255));
        tessFace->rgba_vertices.push_back((uint8_t)(set.color[set.index[i]].b*255));
      }
      numberPoints++;
    }
  }
  if(!set.color.empty())
    tessFace->is_rgba=false;

  tessFace->sizes_triangulated[0]=numberOfTriangleFans;

  tess->face_tessellation.push_back(tessFace);
  return file.add3DTess(tess);
}   

void readTriangleFanSet(X3DTriangleFanSet &set, ticpp::Element *pSet)
{
  string strValue;
  pSet->GetAttribute( "solid", &strValue, false);
  if(strValue=="false")
    set.solid = false;
  pSet->GetAttribute( "fanCount", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      int32_t fanCount;
      stringstream >> fanCount;
      stringstream >> ws;
      set.fanCount.push_back(fanCount);
    }
  }
  if(pSet->FirstChildElement( "Color", false ))
  {
    pSet->FirstChildElement( "Color" )->GetAttribute( "color", &strValue);
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      X3DColor color;
      stringstream >> color.r >> color.g >> color.b;
      stringstream >> c >> ws;
      set.color.push_back(color);
    }
  }
  pSet->FirstChildElement( "Coordinate" )->GetAttribute( "point", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      PRCVector3d point;
      stringstream >> point.x >> point.y >> point.z;
      stringstream >> c >> ws;
      set.point.push_back(point);
    }
  }
}

uint32_t maketessTriangleFanSet(const X3DTriangleFanSet &set, oPRCFile &file)
{
  PRC3DTess *tess = new PRC3DTess;
  PRCTessFace *tessFace = new PRCTessFace;
  tessFace->used_entities_flag=PRC_FACETESSDATA_TriangleFan;
  tessFace->sizes_triangulated.push_back(set.fanCount.size()); // to be filled later
  for(uint32_t i = 0; i < set.point.size(); ++i)
  {
    tess->coordinates.push_back(set.point[i].x);
    tess->coordinates.push_back(set.point[i].y);
    tess->coordinates.push_back(set.point[i].z);
    tess->triangulated_index.push_back(i*3);
    if(!set.color.empty())
    {
      tessFace->rgba_vertices.push_back((uint8_t)(set.color[i].r*255));
      tessFace->rgba_vertices.push_back((uint8_t)(set.color[i].g*255));
      tessFace->rgba_vertices.push_back((uint8_t)(set.color[i].b*255));
    }
  }
  for(size_t i = 0; i < set.fanCount.size(); ++i)
  {
      tessFace->sizes_triangulated.push_back(set.fanCount[i]);
  }
  if(!set.color.empty())
    tessFace->is_rgba=false;

  tess->face_tessellation.push_back(tessFace);
  return file.add3DTess(tess);
}   

void readIndexedLineSet(X3DIndexedLineSet &set, ticpp::Element *pSet)
{
  string strValue;
  pSet->GetAttribute( "coordIndex", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      int32_t index;
      stringstream >> index;
      stringstream >> ws;
      set.index.push_back(index);
    }
  }
  if(pSet->FirstChildElement( "Color", false ))
  {
    pSet->FirstChildElement( "Color" )->GetAttribute( "color", &strValue);
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      X3DColor color;
      stringstream >> color.r >> color.g >> color.b;
      stringstream >> c >> ws;
      set.color.push_back(color);
    }
  }
  pSet->FirstChildElement( "Coordinate" )->GetAttribute( "point", &strValue);
  {
    std::istringstream stringstream(strValue);
    while(!stringstream.eof())
    {
      char c;
      PRCVector3d point;
      stringstream >> point.x >> point.y >> point.z;
      stringstream >> c >> ws;
      set.point.push_back(point);
    }
  }
}

uint32_t maketessIndexedLineSet(const X3DIndexedLineSet &set, oPRCFile &file)
{
  PRC3DWireTess *tess = new PRC3DWireTess();

  for(size_t i = 0; i < set.point.size(); ++i)
  {
    tess->coordinates.push_back(set.point[i].x);
    tess->coordinates.push_back(set.point[i].y);
    tess->coordinates.push_back(set.point[i].z);
  }
  uint32_t numberPoints = 0;
  uint32_t edgeStartIndex = 0;
  tess->wire_indexes.push_back(0); // to be filled later
  for(size_t i = 0; i < set.index.size(); ++i)
  {
    if(set.index[i]==-1)
    {
      tess->wire_indexes[edgeStartIndex] = numberPoints;
      numberPoints = 0;
      edgeStartIndex = tess->wire_indexes.size();
      tess->wire_indexes.push_back(0); // to be filled later
    }
    else
    {
      tess->wire_indexes.push_back(set.index[i]*3);
      if(!set.color.empty()&&numberPoints) // vertex colors do not work, change them into segment colors
      {
        tess->rgba_vertices.push_back((uint8_t)(0.5*(set.color[set.index[i]].r+set.color[set.index[i-1]].r)*255));
        tess->rgba_vertices.push_back((uint8_t)(0.5*(set.color[set.index[i]].g+set.color[set.index[i-1]].g)*255));
        tess->rgba_vertices.push_back((uint8_t)(0.5*(set.color[set.index[i]].b+set.color[set.index[i-1]].b)*255));
      }
      numberPoints++;
    }
  }
  tess->wire_indexes.pop_back(); // remove extra 0
  if(!set.color.empty())
  {
    tess->is_rgba=false;
    tess->is_segment_color=true;
  }

  return file.add3DWireTess(tess);
}   

const uint32_t maketransform(const X3DTransformList &transforms, oPRCFile &file)
{
  if(transforms.empty())
    return m1;
  if(transforms.size()==1)
  {
    PRCCoordinateSystem *coordinateSystem = new PRCCoordinateSystem;
    PRCCartesianTransformation3d *axis_set = new PRCCartesianTransformation3d; 
    const X3DTransform &transform = transforms.front();
    if(!(transform.translation==PRCVector3d(0,0,0)))
    {
      axis_set->behaviour |= PRC_TRANSFORMATION_Translate;
      axis_set->origin = PRCVector3d(transform.translation.x,transform.translation.y,transform.translation.z);
    }
    if(!(transform.scale==PRCVector3d(1,1,1)))
    {
      if((transform.scale.x==transform.scale.y)&&(transform.scale.x==transform.scale.z))
      {
        axis_set->behaviour |= PRC_TRANSFORMATION_Scale;
        axis_set->uniform_scale = transform.scale.x;
      }
      else
      {
        axis_set->behaviour |= PRC_TRANSFORMATION_NonUniformScale;
        axis_set->scale.x = transform.scale.x;
        axis_set->scale.y = transform.scale.y;
        axis_set->scale.z = transform.scale.z;
      }
    }
    coordinateSystem->setAxisSet(axis_set);
    return file.addCoordinateSystem(coordinateSystem);
  }
  // TODO
  return m1;
}

uint32_t makeStyle(const X3DMaterial& material, oPRCFile &file, bool vertexcolor)
{
  if(!vertexcolor)
  {
    PRCMaterialGeneric *materialGeneric = new PRCMaterialGeneric;
    const PRCRgbColor ambient(material.ambientIntensity*material.diffuseColor.r,material.ambientIntensity*material.diffuseColor.g,material.ambientIntensity*material.diffuseColor.b);
    materialGeneric->ambient = file.addRgbColorUnique(ambient);
    if(material.diffuseColor==X3DColor()) // workaround for Adobe viewer
    {
      const PRCRgbColor diffuse(material.emissiveColor.r,material.emissiveColor.g,material.emissiveColor.b);
      materialGeneric->diffuse = file.addRgbColorUnique(diffuse);
      const PRCRgbColor emissive(material.emissiveColor.r/2,material.emissiveColor.g/2,material.emissiveColor.b/2);
      materialGeneric->emissive = file.addRgbColorUnique(emissive);
    } else {
      const PRCRgbColor diffuse(material.diffuseColor.r,material.diffuseColor.g,material.diffuseColor.b);
      materialGeneric->diffuse = file.addRgbColorUnique(diffuse);
      const PRCRgbColor emissive(material.emissiveColor.r,material.emissiveColor.g,material.emissiveColor.b);
      materialGeneric->emissive = file.addRgbColorUnique(emissive);
    }
    const PRCRgbColor diffuse(material.diffuseColor.r,material.diffuseColor.g,material.diffuseColor.b);
    materialGeneric->diffuse = file.addRgbColorUnique(diffuse);
    const PRCRgbColor emissive(material.emissiveColor.r,material.emissiveColor.g,material.emissiveColor.b);
    materialGeneric->emissive = file.addRgbColorUnique(emissive);
    const PRCRgbColor specular(material.specularColor.r,material.specularColor.g,material.specularColor.b);
    materialGeneric->specular = file.addRgbColorUnique(specular);
    materialGeneric->shininess = material.shininess;
    if(material.transparency>0)
    {
      materialGeneric->ambient_alpha  = 1-material.transparency;
      materialGeneric->diffuse_alpha  = 1-material.transparency;
      materialGeneric->emissive_alpha = 1-material.transparency;
      materialGeneric->specular_alpha = 1-material.transparency;
    }
    const uint32_t material_generic_index = file.addMaterialGeneric(materialGeneric);
    PRCStyle *style = new PRCStyle;
    //        style->line_width = lineProperties.linewidthScaleFactor;
    style->line_width = 1;
    style->line_pattern_vpicture_index = 0; //TODO linetype if it gets support in Adobe Reader
    style->is_material = true;
    style->color_material_index = material_generic_index;
    if(material.transparency>0)
    {
      style->is_transparency_defined = true;
      style->transparency = (uint8_t)(material.transparency * 255);
    }
    style->additional = 0;
    return file.addStyle(style);
  }
  else // special material for vertex color, may have to remove it if Adobe fixes viewer
  {
    const PRCRgbColor white(1,1,1);
    const PRCRgbColor gray(1,1,1);
    const uint32_t white_index = file.addRgbColorUnique(white);
    const uint32_t gray_index = file.addRgbColorUnique(gray);
    PRCStyle *style = new PRCStyle;
    style->line_width = 1;
    style->line_pattern_vpicture_index = 0; //TODO linetype
    PRCMaterialGeneric *materialGeneric = new PRCMaterialGeneric;
    materialGeneric->ambient = gray_index;
    materialGeneric->diffuse = white_index;
    materialGeneric->emissive = gray_index;
    materialGeneric->specular = gray_index;
    materialGeneric->shininess = material.shininess;
    if(material.transparency>0)
    {
      materialGeneric->ambient_alpha  = 1-material.transparency;
      materialGeneric->diffuse_alpha  = 1-material.transparency;
      materialGeneric->emissive_alpha = 1-material.transparency;
      materialGeneric->specular_alpha = 1-material.transparency;
    }
    const uint32_t material_generic_index = file.addMaterialGeneric(materialGeneric);
    
    style->is_material = true;
    style->color_material_index = material_generic_index;
    if(material.transparency>0)
    {
      style->is_transparency_defined = true;
      style->transparency = (uint8_t)(material.transparency * 255);
    }
    style->additional = 0;
    return file.addStyle(style);
  }
}

void processElement(ticpp::Element* pElem, X3DTransformList transforms, string DEF, oPRCFile &file, PRCPartDefinition *part_definition)
{
  string strName;
  string strValue;
  pElem->GetValue(&strName);
  
  if(strName=="Transform")
  {
    X3DTransform transform;
    if(pElem->HasAttribute( "translation" ))
    {
      pElem->GetAttribute( "translation", &strValue);
      std::istringstream stringstream(strValue);
      stringstream >> transform.translation.x >> transform.translation.y >> transform.translation.z;
      if(abs(transform.translation.z)<FLT_EPSILON)
        transform.translation.z = 0;
    }
    if(pElem->HasAttribute( "scale" ))
    {
      pElem->GetAttribute( "scale", &strValue);
      std::istringstream stringstream(strValue);
      stringstream >> transform.scale.x >> transform.scale.y >> transform.scale.z;
    }
    if(pElem->HasAttribute( "rotation" ))
    {
      pElem->GetAttribute( "rotation", &strValue);
      std::istringstream stringstream(strValue);
      stringstream >> transform.rotation.axis.x >> transform.rotation.axis.y >> transform.rotation.axis.z >> transform.rotation.angle;
    }
    if(pElem->HasAttribute( "DEF" ))
      pElem->GetAttribute( "DEF", &DEF);

    if(!(transform==X3DTransform()))
    {
      if(!transforms.empty()&&transforms.back().translation==PRCVector3d(0,0,0)&&transforms.back().rotation==X3DRotation()&&transform.scale==PRCVector3d(1,1,1)&&transform.rotation==X3DRotation())
      {
        transforms.back().translation.x=transform.translation.x*transforms.back().scale.x;
        transforms.back().translation.y=transform.translation.y*transforms.back().scale.y;
        transforms.back().translation.z=transform.translation.z*transforms.back().scale.z;
      }
      else
       transforms.push_back(transform);
    }
  }

  ticpp::Iterator<ticpp::Element> child;
  for(child = child.begin(pElem); child != child.end(); child++)
  {
    child->GetValue(&strName);

    if(strName=="Transform"||strName=="Group")
      processElement(child.Get(), transforms, DEF, file, part_definition);

    if(strName=="Viewpoint")
    {
      mat[0][0]=1; mat[0][1]=0; mat[0][2]=0;
      mat[1][0]=0; mat[1][1]=1; mat[1][2]=0;
      mat[2][0]=0; mat[2][1]=0; mat[2][2]=1;
      for(size_t i=0; i<transforms.size(); i++)
      {
//      cout << "transform: " << transforms[i] << endl;
        double rotmat[3][3];
        const double angle = transforms[i].rotation.angle;
        const PRCVector3d &axis = transforms[i].rotation.axis;
        if(axis==PRCVector3d(1,0,0))
        {
          rotmat[0][0] = 1; rotmat[0][1] = 0;          rotmat[0][2] = 0;
          rotmat[1][0] = 0; rotmat[1][1] = cos(angle); rotmat[1][2] = -sin(angle);
          rotmat[2][0] = 0; rotmat[2][1] = sin(angle); rotmat[2][2] =  cos(angle);
        }
        if(axis==PRCVector3d(0,1,0))
        {
          rotmat[0][0] =  cos(angle); rotmat[0][1] = 0; rotmat[0][2] = sin(angle);
          rotmat[1][0] =  0;          rotmat[1][1] = 1; rotmat[1][2] = 0;
          rotmat[2][0] = -sin(angle); rotmat[2][1] = 0; rotmat[2][2] = cos(angle);
        }
        if(axis==PRCVector3d(0,0,1))
        {
          rotmat[0][0] =  cos(angle); rotmat[0][1] = -sin(angle); rotmat[0][2] = 0;
          rotmat[1][0] =  sin(angle); rotmat[1][1] =  cos(angle); rotmat[1][2] = 0;
          rotmat[2][0] =  0;          rotmat[2][1] =  0;          rotmat[2][2] = 1;
        }
        double newmat[3][3];
        for(size_t k=0; k<3; k++)
          for(size_t l=0; l<3; l++)
            newmat[k][l] = mat[k][0]*rotmat[0][l]+mat[k][1]*rotmat[1][l]+mat[k][2]*rotmat[2][l];
        for(size_t k=0; k<3; k++)
          for(size_t l=0; l<3; l++)
            mat[k][l] = newmat[k][l];
      }
    }

    if(strName=="Shape"&&child->FirstChildElement( "Appearance", false )&&child->FirstChildElement( "Appearance" )->FirstChildElement( "Material", false ))
    {
//      cout << "Shape " << endl; 

      X3DMaterial material;
      if(child->FirstChildElement( "Appearance" )->FirstChildElement( "Material", false ))
      {
        ticpp::Element *pMaterial;
        pMaterial = child->FirstChildElement( "Appearance" )->FirstChildElement( "Material" );
        pMaterial->GetAttribute( "ambientIntensity", &material.ambientIntensity);
        pMaterial->GetAttribute( "shininess", &material.shininess);
        pMaterial->GetAttribute( "transparency", &material.transparency);
        pMaterial->GetAttribute( "diffuseColor", &strValue);  { std::istringstream stringstream(strValue); stringstream >> material.diffuseColor; }
        pMaterial->GetAttribute( "emissiveColor", &strValue); { std::istringstream stringstream(strValue); stringstream >> material.emissiveColor; }
        pMaterial->GetAttribute( "specularColor", &strValue); { std::istringstream stringstream(strValue); stringstream >> material.specularColor; }
//        if(!(material==X3DMaterial()))
//          cout << material << endl;
      }

//    X3DLineProperties lineProperties;
//    if(child->FirstChildElement( "Appearance" )->FirstChildElement( "LineProperties", false ))
//    {
//      ticpp::Element *pLineProperties;
//      pLineProperties = child->FirstChildElement( "Appearance" )->FirstChildElement( "LineProperties" );
//      pLineProperties->GetAttribute( "linetype", &lineProperties.linetype, false);
//      pLineProperties->GetAttribute( "linewidthScaleFactor", &lineProperties.linewidthScaleFactor, false);
//      if(!(lineProperties==X3DLineProperties()))
//        cout << lineProperties << endl;
//    }

      if(child->FirstChildElement( "IndexedTriangleStripSet", false ))
      {
        X3DIndexedTriangleSet set;
        readIndexedTriangleSet(set, child->FirstChildElement( "IndexedTriangleStripSet" ));

        // create material
        const uint32_t style_material_index = makeStyle(material, file, !set.color.empty());

        const uint32_t tess_index = maketessIndexedTriangleStripSet(set, file);
        const uint32_t transform_index = maketransform(transforms, file);;

        PRCPolyBrepModel *polyBrepModel = new PRCPolyBrepModel("TriangleStrip"+DEF);
        polyBrepModel->index_of_line_style = style_material_index;
        polyBrepModel->index_tessellation = tess_index;
        polyBrepModel->index_local_coordinate_system = transform_index;
        if(!set.color.empty()&&material.transparency>0)
        {
          transparentMeshes.push_back(polyBrepModel);
          transparencies.push_back(material.transparency);
        } 
        part_definition->addPolyBrepModel(polyBrepModel);
      }
      
      if(child->FirstChildElement( "TriangleFanSet", false ))
      {
        X3DTriangleFanSet set;
        readTriangleFanSet(set, child->FirstChildElement( "TriangleFanSet" ));

        // create material
        const uint32_t style_material_index = makeStyle(material, file, !set.color.empty());

        const uint32_t tess_index = maketessTriangleFanSet(set, file);
        const uint32_t transform_index = maketransform(transforms, file);;

        PRCPolyBrepModel *polyBrepModel = new PRCPolyBrepModel("TriangleFan"+DEF);
        polyBrepModel->index_of_line_style = style_material_index;
        polyBrepModel->index_tessellation = tess_index;
        polyBrepModel->index_local_coordinate_system = transform_index;
        part_definition->addPolyBrepModel(polyBrepModel);
      }
      
      if(child->FirstChildElement( "IndexedLineSet", false ))
      {
        X3DIndexedLineSet set;
        readIndexedLineSet(set, child->FirstChildElement( "IndexedLineSet" ));

        // create material
        uint32_t style_material_index = m1;
        if(set.color.empty())
        {
          const PRCRgbColor color(material.emissiveColor.r,material.emissiveColor.g,material.emissiveColor.b);
          const uint32_t color_index = file.addRgbColorUnique(color);
          PRCStyle *style = new PRCStyle("IndexedLineSetStyle");
          style->line_width = 1;
          style->line_pattern_vpicture_index = 0; //TODO linetype
          style->is_material = false;
          style->color_material_index = color_index;
          style->additional = 0;
          style_material_index = file.addStyle(style);
        }
        else // special material for vertex color, may have to remove it if Adobe fixes viewer
        {
          const PRCRgbColor color(1,1,1);
          const uint32_t color_index = file.addRgbColorUnique(color);
          PRCStyle *style = new PRCStyle("VertexColorLineSetStyle");
          style->line_width = 1;
          style->line_pattern_vpicture_index = 0; //TODO linetype
          style->is_material = false;
          style->color_material_index = color_index;
          style->additional = 0;
          style_material_index = file.addStyle(style);
        }

        const uint32_t tess_index = maketessIndexedLineSet(set, file);
        const uint32_t transform_index = maketransform(transforms, file);;

        PRCPolyWire *polyWire = new PRCPolyWire("Line"+DEF);
        polyWire->index_of_line_style = style_material_index;
        polyWire->index_tessellation = tess_index;
        polyWire->index_local_coordinate_system = transform_index;
        part_definition->addPolyWire(polyWire);

      }
    }

    if(strName=="Billboard")
    {
      child->FirstChildElement( "Transform" )->FirstChildElement( "Shape" )->FirstChildElement( "Box" )->GetAttribute( "size", &strValue);
      double x,y,z;
      {
        std::istringstream stringstream(strValue);
        stringstream >> x >> y >> z;
      }
      child->FirstChildElement( "Transform" )->FirstChildElement( "Shape" )->FirstChildElement( "Appearance" )->FirstChildElement( "PixelTexture" )->GetAttribute( "image", &strValue);
      uint32_t width, height, components;
      vector<uint32_t> pixels;
      {
        std::istringstream stringstream(strValue);
        stringstream >> width >> height >> components;
        stringstream >> hex;
        while(!stringstream.eof())
        {
          uint32_t pixel;
          stringstream >> pixel;
          stringstream >> ws;
          pixels.push_back(pixel);
        }
      }

      // create material
      uint32_t style_texture_index = m1;
      {
        if(black_material_generic_index==m1) {
          PRCMaterialGeneric *materialGeneric = new PRCMaterialGeneric("BoxMaterial");
          const PRCRgbColor black(0,0,0);
          const uint32_t black_index = file.addRgbColorUnique(black);
          materialGeneric->ambient = black_index;
          materialGeneric->diffuse = black_index;
          materialGeneric->emissive = black_index;
          materialGeneric->specular = black_index;
          materialGeneric->shininess = 0;
          black_material_generic_index = file.addMaterialGeneric(materialGeneric);
        }
        uint8_t *picture = new uint8_t[width*height*4];
        for(size_t i=0; i<pixels.size(); i++)
        {
          picture[i*4+0]=(pixels[i]&0xFF000000)>>24;
          picture[i*4+1]=(pixels[i]&0xFF0000)>>16;
          picture[i*4+2]=(pixels[i]&0xFF00)>>8;
          picture[i*4+3]=(pixels[i]&0xFF);
        }
        const uint32_t picture_index = file.addPicture(KEPRCPicture_BITMAP_RGBA_BYTE,width*height*4,picture, width,height,"box_picture");
        delete[] picture;

        PRCTextureDefinition *textureDefinition = new PRCTextureDefinition("box_texture_definition");
        textureDefinition->picture_index = picture_index;
        textureDefinition->texture_function = KEPRCTextureFunction_Replace;
        textureDefinition->texture_mapping_attribute_components = PRC_TEXTURE_MAPPING_COMPONENTS_RGBA;
        const uint32_t texture_definition_index = file.addTextureDefinition(textureDefinition);
        PRCTextureApplication *textureApplication = new PRCTextureApplication("box_texture_application");
        textureApplication->material_generic_index = black_material_generic_index;
        textureApplication->texture_definition_index = texture_definition_index;
        const uint32_t texture_application_index = file.addTextureApplication(textureApplication);
        PRCStyle *styleTexture = new PRCStyle("style_texture");
        styleTexture->line_pattern_vpicture_index = 0;
        styleTexture->is_material = true;
        styleTexture->color_material_index = texture_application_index;
        style_texture_index = file.addStyle(styleTexture);
      }

      uint32_t tess_index = m1;
      {
        PRC3DTess *tess = new PRC3DTess;
        PRCTessFace *tessFace = new PRCTessFace;
        tessFace->used_entities_flag=PRC_FACETESSDATA_TriangleFanTextured;
        tess->coordinates.push_back(-x/2+0); // 0
        tess->coordinates.push_back(-y/2+0);
        tess->coordinates.push_back(0);
        tess->coordinates.push_back(-x/2+x); // 1
        tess->coordinates.push_back(-y/2+0);
        tess->coordinates.push_back(0);
        tess->coordinates.push_back(-x/2+x); // 2
        tess->coordinates.push_back(-y/2+y);
        tess->coordinates.push_back(0);
        tess->coordinates.push_back(-x/2+0); // 3
        tess->coordinates.push_back(-y/2+y);
        tess->coordinates.push_back(0);
        tess->texture_coordinate.push_back(0); // 0
        tess->texture_coordinate.push_back(0);
        tess->texture_coordinate.push_back(1); // 1
        tess->texture_coordinate.push_back(0);
        tess->texture_coordinate.push_back(1); // 2
        tess->texture_coordinate.push_back(1);
        tess->texture_coordinate.push_back(0); // 3
        tess->texture_coordinate.push_back(1);
        tess->triangulated_index.push_back(0*2);
        tess->triangulated_index.push_back(0*3);
        tess->triangulated_index.push_back(1*2);
        tess->triangulated_index.push_back(1*3);
        tess->triangulated_index.push_back(2*2);
        tess->triangulated_index.push_back(2*3);
        tess->triangulated_index.push_back(3*2);
        tess->triangulated_index.push_back(3*3);
        tessFace->sizes_triangulated.push_back(1);
        tessFace->sizes_triangulated.push_back(4);
        tessFace->number_of_texture_coordinate_indexes = 1;
        tess->face_tessellation.push_back(tessFace);
        tess_index = file.add3DTess(tess);
      }

      const uint32_t transform_index = maketransform(transforms, file);;

      PRCPolyBrepModel *polyBrepModel = new PRCPolyBrepModel("Billboard");
      polyBrepModel->index_of_line_style = style_texture_index;
      polyBrepModel->index_tessellation = tess_index;
      polyBrepModel->index_local_coordinate_system = transform_index;
      polyBrepModel->is_closed = true;
      part_definition->addPolyBrepModel(polyBrepModel);
    }

    if(strName=="ProtoInstance")
    {
      child->GetAttribute( "name", &strName);
      child->FirstChildElement( "fieldValue" )->GetAttribute( "value", &strValue );
      std::istringstream stringstream(strValue);
      X3DColor emissiveColor;
      stringstream >> emissiveColor.r >> emissiveColor.g >> emissiveColor.b;

      // create material

      const uint32_t transform_index = maketransform(transforms, file);;

      if(protos[strName].type == "IndexedLineSet")
      {
        const PRCRgbColor color(emissiveColor.r,emissiveColor.g,emissiveColor.b);
        const uint32_t color_index = file.addRgbColorUnique(color);
        PRCStyle *style = new PRCStyle;
        style->line_width = 1;
        style->line_pattern_vpicture_index = 0;
        style->is_material = false;
        style->color_material_index = color_index;
        style->additional = 0;
        const uint32_t style_material_index = file.addStyle(style);

        PRCPolyWire *polyWire = new PRCPolyWire("ProtoInstance");
        polyWire->index_of_line_style = style_material_index;
        polyWire->index_tessellation = protos[strName].tess_index;
        polyWire->index_local_coordinate_system = transform_index;
        part_definition->addPolyWire(polyWire);
      }
      else
      {
        PRCMaterialGeneric *materialGeneric = new PRCMaterialGeneric;
        const PRCRgbColor ambient(0,0,0);
        materialGeneric->ambient = file.addRgbColorUnique(ambient);
        const PRCRgbColor diffuse(0,0,0);
//      const PRCRgbColor diffuse(emissiveColor.r,emissiveColor.g,emissiveColor.b);
        materialGeneric->diffuse = file.addRgbColorUnique(diffuse);
        const PRCRgbColor emissive(emissiveColor.r,emissiveColor.g,emissiveColor.b);
//      const PRCRgbColor emissive(emissiveColor.r/2,emissiveColor.g/2,emissiveColor.b/2);
        materialGeneric->emissive = file.addRgbColorUnique(emissive);
        const PRCRgbColor specular(0,0,0);
        materialGeneric->specular = file.addRgbColorUnique(specular);
        const uint32_t material_generic_index = file.addMaterialGeneric(materialGeneric);
        PRCStyle *style = new PRCStyle;
        style->line_width = 1;
        style->line_pattern_vpicture_index = 0;
        style->is_material = true;
        style->color_material_index = material_generic_index;
        style->additional = 0;
        const uint32_t style_material_index = file.addStyle(style);
      
        PRCPolyBrepModel *polyBrepModel = new PRCPolyBrepModel("ProtoInstance");
        polyBrepModel->index_of_line_style = style_material_index;
        polyBrepModel->index_tessellation = protos[strName].tess_index;
        polyBrepModel->index_local_coordinate_system = transform_index;
        polyBrepModel->is_closed = true;
        part_definition->addPolyBrepModel(polyBrepModel);
      }
    }

  }
}

